% Demo - introduction to membership functions.

% (c) Robert Babuska, March 1994

echo off
% Reset things before we get started.
clear
clc;clf;
format
echo on
%       MEMBERSHIP FUNCTIONS
%
% This part introduces the basic types of analytically defined membership
% functions and explains how they can be handled in MATLAB. Analytical
% membership functions are useful for defining fuzzy sets on continuous
% domains. Function MGRADE calculates the membership grade(s) of domain
% element(s) in the fuzzy set(s).

% Press any key to continue...
pause
clc
% Function MGRADE has two arguments, crisp scalar (vector) and
% membership function(s) definition vector (matrix), MGRADE(X,MFS).
% Membership functions are defined in MFS rows, with the following format:
%       MFS = [MFTYPE, P1, P2, P3, P4]
% where integer MFTYPE determines the type of the membership function
%       0 ... one everywhere,        1 ... Trapezoidal
%       2 ... Exponential,           3 ... Sigmoidal

% Press any key to continue...
pause
clc
% The remaining four parameters P1...P4 define the membership function
% position and shape as follows (Ax is the membership grade of crisp number X
% in fuzzy set A):
%       Ax = 0                  for x <= P1 or  x >= P2
%            1                  for x >= P2 and x <= P3
%            in (0,1)           for x >  P1 and x <  P2 or x >  P3 and x <  P4
%
% The shape of the membership function curve for X between P1 and P2
% and between P3 and P4 is determined by membership function type,
% i.e. trapezoidal, exponential or sigmoidal.

% Press any key to continue...
pause
clc
% In most applications the particular shape is not critical, much more
% important is for instance the overlap between neighboring functions. However,
% in certain cases, the choice of membership function shape can be useful. As
% and example, let us mention parametrization of point-wise defined fuzzy
% clusters obtained from clustering algorithms.

% Press any key to continue...
pause
clc
% The examples for different membership function types are given below. First
% we define a universe of discourse:
x = -1:0.01:1;

%       TRAPEZOIDAL membership function, MFTYPE = 1

mftrap = [1 -0.5 -0.05 0.05 0.5];
Ax = mgrade(x,mftrap);

plot(x,Ax);
title('Trapezoidal membership function');
xlabel('X'); ylabel('Membership grade');

% Press any key to continue...
pause
clc

%       EXPONENTIAL Membership Function, MFTYPE = 2

mfexp=[2 0.0 0.5 0.5 1.0];
Ax = mgrade(x,mfexp);

plot(x,Ax);
title('Exponential membership function');
xlabel('X'); ylabel('Membership grade');

% Press any key to continue...
pause
clc

%       SIGMOIDAL Membership Function, MFTYPE = 3

mfsigm=[3 -0.7 -0.1 0.2 0.8];
Ax = mgrade(x,mfsigm);

plot(x,Ax);
title('Sigmoidal membership function');
xlabel('X'); ylabel('Membership grade');

% Press any key to continue...
pause
clc
% MFS can be also a n by 5 matrix with each row defining one membership function
% (fuzzy set). In this case a corresponding vector of membership grades is
% returned by MGRADE. This is exactly function of the fuzzifier block in fuzzy
% controllers. In the following example three fuzzy sets for linguistic terms
% describing temperature in a room (COLD, OK, HOT) are defined.

x = 10:0.1:35;
A = [1 10 10 17 22;
     1 17 20 22 27;
     1 22 26 35 35];
Ax = mgrade(x,A);

plot(x,flipud(Ax));
title(' COLD           OK                  HOT');
xlabel('Temperature [centigrade]'); ylabel('Membership degree');

% Press any key to continue...
pause
clc
% For practical applications (fuzzy control, decision-making, ...) it is
% useful to have a means for graphical design of membership functions.
% Commercial SW packages (TILShell, FIDE, A-B FLEX, ...) offer for this
% purpose special graphical editors. In Fuzzy Toolbox a similar utility,
% based on MATLAB Handle Graphic, is available.

% Press any key to continue...
pause
clc
% A pair of functions, MFSET and MFGET are used for membership function
% design. MFSET installs the editor in a specified graphical window (figure)
% and sets initial parameters. Then the mouse is used for changing the
% function shapes and positions. MFGET reads the parameters from the editor in
% a specified figure.

% Press any key to continue...
pause
clc

% You can use the mouse in order to change the membership functions. With
% the four sliders in the corners of the figure, change the shape of the
% selected function or of all functions simultaneously, if no function is
% selected. Every membership function can be selected and deselected by clicking
% the curve of the function. The purpose of the wide slider in the upper part of
% the figure is two-fold. When no function is selected, it sets the crisp input
% values (red vertical line in the graph). If one of the membership functions is
% selected, the slider is used to move the function along the domain.

% Press any key to continue...
pause
clc

% To get the parameters of membership functions into the MATLAB workspace, we
% use function MFGET(FIGNO, 'MFS'). FIGNO is the number of figure and 'MFS' is
% a property saying that membership functions should be get from the editor.
% Other properties can be read as well. For details type HELP MFGET.

% Press any key to quit...
pause
clc
echo off; clc
