function [rls, mfs, inputs, outputs, nlab, labels] = txt2rls(filename,str);
% TXT2RLS parses a rule base given in a text form and codes the rules
% in a matrix. [RLS,MFS,INPUTS,OUTPUTS,NLAB,LABELS]  = TXT2RLS(FILENAME)
%
% Input:      FILENAME  ...  name of the text file or a text matrix
%                            containing the rule-base
% Outputs :   RLS       ...  rule-base matrix
%             MFS       ...  membership functions matrix
%             INPUTS    ...  input labels
%             OUTPUTS   ...  output labels
%             NLAB      ...  number of labels for all inputs and outputs
%                            i.e. [i1; i2; i3; ... o1; o2; ...]
%             LABELS    ...  input and output labels concatenated in one
%                            matrix, i.e. [Input1_Label1; Input1_Label2;
%                                                       ...
%                                          Input2_Label1; Input2_Label2;
%                                                       ...
%                                          Output1_Label1;Output1_Label2;
%                                                       ...              ]
%
% [RLS,MFS,INPUTS,OUTPUTS,NLAB,LABELS]  = TXT2RLS(STRING,'string') reads
% the rule base from a text matrix STRING. 
%
% Input and output identifiers and fuzzy set labels for each input and
% output must be defined. If-Then rules can contain logical operators
% AND and NOT (NOT is applicable only in rule premises). For a sample
% rule-base see the SAMPLE.RB file.
%
% Rule base items are:
%       directives             - #input, #output, #rules, #mf (optional)
%       keywords               - If, Then, Is, And, Not, comma (,)
%       rule terminator        - ;
%       input and output names - user-defined identifiers
%       fuzzy set labels       - user-defined identifiers
%       rule weights           - real numbers
%       membership fcns defs   - five-tuples of real numbers
%       singletons             - alternatively, rule conclusions can be
%                                defined as singletons, i.e. real numbers.
%                                In such case, no labels are defined after
%                                the #output directive.
%
% Identifiers (input and output names and fuzzy set labels) can consist
% of multiple words (e.g. error change) but cannot contain keywords,
% directives and semicolon. Identifiers are case-sensitive, keywords and
% directives case-insensitive. Input and output names follow immediately
% after their respective directives. Each fuzzy set label must be written
% on a separate line, i.e.:
%                        #input error change
%                               Big Positive
%                               Medium Positive
%                                    ...
%
% A rule is a text fragment beginning with "If" and ending with a
% semicolon (;). A rule must begin on a new line and can span
% multiple lines. Every rule can have a weight which is a real number
% preceding the rule on the same or separate line. Example:
%
%       0.85 IF error is POSITIVE BIG and output is POSITIVE
%            THEN control is NEGATIVE BIG;
%
% Membership functions definitions can be placed anywhere in the text.
% Every membership function is defined by a five-tuple [type a b c d].
% For details on types and parameters see MGRADE. Example:
%
%       #input error change
%               Big Positive	#mf 1 0.0 0.8 1.0 1.0
%               Medium Positive #mf 1 0.0 0.4 0.5 0.9
%
% Any text following the '%' character until CR/LF is a comment. The order
% of membership functions, input, output definitions and rules is arbitrary.
%
% See also SAMPLE.RB, RLS2TXT, MFSET, MFGET, INFER, MGRADE

% Copyright (c) Robert Babuska, 1993-94

% directives and keywords
inpstr = '#input'; outstr  = '#output'; rulstr  = '#rules';
mfdstr = '#mf';

IfStr   = ' if '; ThenStr = ' then '; IsStr   = ' is ';
AndStr  = ' and '; NotStr  = 'not '; RuleEnd = ';';
CSepStr = ','; Comment = '%'; LF = setstr(10); 
CR = setstr(13); TAB = setstr(9);

if nargin > 1,						% RB in string
  if strcmp(str,'string'),
    A = filename;
    if min(size(A)) > 1,				% text matrix
      A = [A LF*ones(size(A,1),1)]';
    end;
    A = A(:).';						% make always row
  else 
    error(['Unknown option ' str]); 
  end; 
else							% RB in file
  if ~any(filename=='.'), filename = [filename '.rb']; end;
  fid=fopen(filename);                    		% open filename
  if fid == -1, error(['Error opening file ' filename]); end;
  A = fscanf(fid,'%c');
  fclose(fid);
end;

tabs = A==TAB;                                          % remove tabulators
A(tabs) = setstr(tabs(tabs==1)*' ');                    % and multiple blanks
ind = find(A == ' ');
li = length(ind);
A(ind(find(ind(2:li)-ind(1:li-1) == 1))) = [];
						
if any(A==LF), NewLine = LF; ToRemove = CR; 		% remove CR or LF
else NewLine = CR; ToRemove = LF;  end;
A(A==ToRemove) = [];                                    % remove CR's or LF's

ind = find(A==NewLine);	                                % remove leading blanks
if isempty(ind), 
  error('Rule base syntax error (no CR''s or LF''s found)'); 
end;
ind(ind == length(A)) = [];
A(ind(A(ind+1)==' ')+1)=[];
A = [NewLine A NewLine];
lf = find(A == NewLine);                                 % remove comments
C  = find(A == '%');
LFF = lf'*ones(1,length(C));
CC  = ones(length(lf),1)*C;
CLF = lf(sum((CC>LFF))+1);
for i = length(C) : -1 : 1,
  A(C(i):CLF(i)-1) = [];
end;
ind = find(A == NewLine);                                    	% remove multiple NewLine codes
li = length(ind);
A(ind(find(ind(2:li)-ind(1:li-1) == 1))) = [];
ind = find(A==NewLine);                              		% remove trailing blanks
ind(1) = [];
A(ind(A(ind-1)==' ')-1)=[];
A(1) = [];

dirs = find(A == '#'); ld = length(dirs);           		% put NewLine before all '#'
dirs = dirs(2:ld);
if ~all(A(dirs-1) == NewLine | A(dirs-1) == ' ')
  error('Syntax error - space or CR/LF missing before a directive.'); 
end;
A(dirs-1) = setstr(NewLine*ones(1,ld-1));			

csep = find(A==CSepStr)+1;			% find indices following output separators
if length(csep) > 0,
  csep = csep(A(csep)==' ' | A(csep) == NewLine);
  A(csep) = [];                 		% remove blanks and NewLine codes following separators
end;

A = [A '#' NewLine];            % add '#' and NewLine at the end of file
a = lower(A);			
lf = find(a == NewLine);        % find NewLine
inps = findstr(a,inpstr);       % find input directive positions
ninps = length(inps);           % find # of inputs
nisets = zeros(ninps,1);        % prepare vector for numbers of input FS
outs = findstr(a,outstr);       % find output directive positions
nouts = length(outs);           % find # of outputs
nosets = zeros(nouts,1);        % prepare vector for numbers of output FS
rules = findstr(a,rulstr);      % find rules directive position
mfd = findstr(a,mfdstr);        % find mfs directive position
nmfd = length(mfd);             % find # of mfs directives
mfs  = [];                      % prepare empty MF matrix

% process input directives

inputs = [];
mfinp = [];
for i = 1 : ninps,                      % for all input directives
  fpos = inps(i) + length(inpstr) + 1;  % calculate position after the directive
  clf = lf(lf > fpos);                  % find next NewLine
  if A(fpos) == '#', 
    error('Input name missing');
  end;        
  inputs=str2mat(inputs,A(fpos:clf-1)); % save input name
  while 1,                              % get input labels
    fpos = clf(1) + 1;
    if ~isempty(mfd),
       if any(fpos==mfd), 			% mf directive encountered
         clf = lf(lf >= fpos);             % find next NewLine
         fpos = clf(1) + 1;		% skip it
       end;
    end;  
    if A(fpos) == '#', break; end;  	% next directive found (other than mf)
    clf = lf(lf > fpos);                % find next NewLine
    mfinp=str2mat(mfinp,A(fpos:clf-1)); % save label
    nisets(i) = nisets(i) + 1;          % input label count
  end;
  if nisets(i) == 0, 
    error(['No labels defined for input "' inputs(i+1,:) '".']);
  end;
end;
inputs = inputs(2:ninps+1,:);
mfinp = mfinp(2:sum(nisets)+1,:);

% process output directives

outputs = [];
mfout = [];
for i = 1 : nouts,                      % for all outputs
  fpos = outs(i) + length(outstr) + 1;  % calculate position after the directive
  clf = lf(lf > fpos);                  % find next NewLine
  if A(fpos) == '#', 
    error('Output name missing');
  end;        
  outputs=str2mat(outputs,A(fpos:clf-1)); % save output name
  while 1,                              % get input labels
    fpos = clf(1) + 1;
    if ~isempty(mfd),
       if any(fpos==mfd), 			% mf directive encountered
          clf = lf(lf >= fpos);             % find next NewLine
          fpos = clf(1) + 1;		% skip it
       end;
    end;   
    if A(fpos) == '#', break; end;      % next directive found (other than mf)
    clf = lf(lf > fpos);                % find next NewLine
    mfout=str2mat(mfout,A(fpos:clf-1)); % save label
    nosets(i) = nosets(i) + 1;          % output label count
  end;
end;
outputs = outputs(2:nouts+1,:);
mfout = mfout(2:sum(nosets)+1,:);

% process rules

nlab   = [nisets; nosets];
clab   = [0 cumsum(nlab(1:ninps+nouts-1))'];
nrules = sum(a(rules:length(a)) == RuleEnd);    % find number of rules
fpos = rules(1) + 7;				% set pointer
rls = zeros(nrules,ninps+nouts);                % initialize rule matrix
weights = ones(nrules,1);

for k = 1 : nrules,                     % for all rules
  s = [];
  while 1                               % read rule
    clf = lf(lf > fpos);                % find next NewLine
    s = [s ' ' A(fpos:clf-1)];
    fpos = clf(1) + 1;
    ls = find(s == RuleEnd) - 1;
    if length(ls) > 0; break; end;	% rule end found
  end;
  s = s(1:ls);
  if s(ls) == ' ',
    s = s(1:ls-1); ls = ls - 1;
  end;
  
  sLower = [lower(s(1:ls)) ' '];                   % convert to lower-case
  Ifpos = findstr(sLower,IfStr);                   % find "if" position
  Thenpos = findstr(sLower,ThenStr);               % find "then" position
  if length(Ifpos) ~= 1 | length(Thenpos) ~= 1,
      error(['Syntax error in rule ' num2str(k) '.']);
  end;
  w = sscanf(s(1:Ifpos-1),'%f');                   % read rule weight
  if all(size(w)), weights(k) = w; end;            % store rule weight
  ante = s(Ifpos+4:Thenpos-1);                     % get rule antecedent
  cons = s(Thenpos+6:ls);                          % get rule consequent
  aLower = sLower(Ifpos+4:Thenpos-1);              % lower-case
  cLower = sLower(Thenpos+6:ls);                   % lower-case

% antecedents
  if length(aLower) == 0, nis = 0; 
  else
    Is = findstr(aLower,IsStr); nis = length(Is);         % find "Is's"
    And = [-4 findstr(aLower,AndStr) length(ante)+1];     % find "And's"
  end;

  lin = size(inputs,2);
  for j = 1 : nis,
    ii =  ante(And(j)+5:Is(j)-1); lii = Is(j) - And(j)-5;
    if lii > lin,
      error(['Invalid input "' ii '" in rule ' num2str(k) '.']);
    end;
    mfi = ante(Is(j)+4:And(j+1)-1); lmfi = And(j+1) - Is(j) - 4;
    signum = 1;
    if lmfi > 3, if lower(mfi(1:4)) == NotStr,
      signum = -1; mfi = mfi(5:lmfi); lmfi = lmfi - 4;
    end; end;
    InputNo = 0;
    for i = 1 : ninps
      if (all(inputs(i,lii+1:lin) == ' ') | lii == lin) & ii == inputs(i,1:lii), 
        InputNo = i; break; 
      end;
    end;
    if InputNo == 0,
      error(['Invalid input "' ii '" in rule ' num2str(k) '.']);
    end;
    MfNo = 0;
    for i = 1 : nisets(InputNo)
      mfind = sum(nisets(1:InputNo-1)) + i;
      if length(mfinp(mfind,:)) < lmfi,
        error(['Invalid label "' mfi '" for input "' ii '" in rule ' num2str(k) '.']);
      end;
      if mfi == mfinp(mfind,1:lmfi),
        MfNo = signum*(i + clab(InputNo)); break;
      end;
    end;
    if MfNo == 0,
      error(['Invalid label "' mfi '" for input "' ii '" in rule ' num2str(k) '.']);
    end;
    rls(k,InputNo) = MfNo;
  end;

% consequents
  if length(cLower) == 0, nis = 0;
  else 
    Is = findstr(cLower,IsStr); nis = length(Is);
    And = [0 find(cLower==CSepStr) length(cons)+1];
  end;

  lin = size(outputs,2);
  for j = 1 : nis,
    ii =  cons(And(j)+1:Is(j)-1); lii = Is(j) - And(j)-1;
    if lii > lin,
      error(['Invalid output "' ii '" in rule ' num2str(k) '.']);
    end;
    mfi = cons(Is(j)+4:And(j+1)-1); lmfi = And(j+1) - Is(j) - 4;
    signumo = 1;
    if lmfi > 3, if lower(mfi(1:4)) == NotStr,
      signumo = -1; mfi = mfi(5:lmfi); lmfi = lmfi - 4;
    end; end;
    OutputNo = 0;
    for i = 1 : nouts
      if (all(outputs(i,lii+1:lin) == ' ') | lii == lin) & ii == outputs(i,1:lii), 
        OutputNo = i; break; 
      end;
    end;
    if OutputNo == 0,
      error(['Invalid output "' ii '" in rule ' num2str(k) '.']);
    end;
    MfNo = 0;

    if nosets(OutputNo) == 0,                    % Sugeno
      w = sscanf(mfi,'%f');
      if all(size(w)),
        rls(k,ninps+OutputNo) = w; 
      else
        error(['Invalid numerical consequent "' mfi '" for output "' ii '" in rule ' num2str(k) '.']);
      end;
    else                                         % Mamdani
      for i = 1 : nosets(OutputNo)
        mfind = sum(nosets(1:OutputNo-1)) + i;
        if length(mfout(mfind,:)) < lmfi,
          error(['Invalid label "' mfi '" for output "' ii '" in rule ' num2str(k) '.']);
        end;
        if mfi == mfout(mfind,1:lmfi),
          MfNo = signumo*(i + clab(ninps+OutputNo)); break;
        end;
      end;
      if MfNo == 0,
        error(['Invalid label "' mfi '" for output "' ii '" in rule ' num2str(k) '.']);
      end;
      rls(k,ninps+OutputNo) = MfNo;
    end;
  end;
end;

labels = str2mat(mfinp,mfout);
rls = [weights rls];

% process mfs directives

for i = 1 : nmfd,                       % for all mfs directives
  fpos = mfd(i) + length(mfdstr) + 1;   % calculate position after the directive
  clf = lf(lf >= fpos);                 % find next NewLine
  while 1,                              % get input labels
    mf = sscanf(a(fpos:clf-1),'%f')';   % read mfs parameters
    if length(mf) ~= 5,                 % incorrect mf definition, skip rest
	break; 
    end;     
    mfs=[mfs; mf];
    fpos = clf(1) + 1;
    clf = lf(lf > fpos);                % find next NewLine
  end;
end;

